from __future__ import print_function
# -*- coding: utf-8 -*-
# Copyright 2015 Wu Jin <wujin_astro@pku.edu.cn>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''
ccd product
vim:set foldmethod=indent
'''
'''Update history
2016-01-27 add the license
'''
'''Bugs or Cautions:
'''

import astropy.io.fits as pyfits
import astropy
import numpy as np
from numpy.polynomial import Legendre
import matplotlib.pyplot as plt
from . import LACosmic as la
import scipy.interpolate as interp
import os
import ipdb
import sys
from ...utils import messager
from ...utils import macro
from ...math import astroMath
import ctypes
from ctypes import c_int, c_double, c_short

so = ctypes.CDLL(os.path.join(os.path.dirname(__file__), "lib", "imageUtils.so"))
grow_c = so.grow
c_int_p = ctypes.POINTER(c_int)
c_short_p = ctypes.POINTER(c_short)
c_double_p = ctypes.POINTER(c_double)


''' something about pyfits
                  ny      nx
    we will get (nrows, ncolumns) array if we read fits file form pyfits, we called the [nx, ny] index a nature one because in ds9, the x index is the column index adn the y index is the row index.
    we will not do any transepose any more, next time when we deal with data read from pyfits, we just make the call like data[ny, nx], remember it!
'''

#!! check ok
def getx1x2y1y2(string):
    """get numbers from section like [1:1024,1:1028]"""
    temp = string.strip()
    assert temp[0]=='[' and temp[-1]==']'
    temp = temp[1:-1].split(',')
    x1, x2 = [int(each) for each in temp[0].split(':')]
    y1, y2 = [int(each) for each in temp[1].split(':')]
    return x1, x2, y1, y2

#!! check ok
def overscanAndTrim(imageFile, configs={}):#<==
    '''sub overscan and do trim using header imformation
    Parameters
    -------------------
    image : `str` or `astropy.io.fits.hdu.hdulist.HDUList`
        input image
    configs : `dict`, optional
        fit : weather to fit the overscan region
        subDirection : which way to (fit) and sub the overscan region
        headerInfoFrame : {'all', "every"}
            which frame to the the section information
    Returns
    ------------------
    type : `astropy.io.fits.hdu.hdulist.HDUList`

    '''#==>
    #!! format input image
    obj = pyfits.open(imageFile)
    primaryHDU = [pyfits.PrimaryHDU(None, obj[0].header)]
    secondHDU = [pyfits.ImageHDU(obj[0].data, pyfits.Header())]
    hduList = pyfits.HDUList(primaryHDU + secondHDU)
    exec(messager.initLoggerMacro)

    #region: setups
    subDirection = configs.get('subDirection', 'row')
    if subDirection not in ['row', 'column']:
        raise logger.Exception('Error subDirection')
    N = len(hduList)
    # fit for the bias in the overscan region
    fit = configs.get('fit', False)

    mainHeader = hduList[0].header
    bx1, bx2, by1, by2 = getx1x2y1y2(mainHeader['BIASSEC'])
    assert mainHeader['TRIMSEC']==mainHeader['DATASEC']
    dx1, dx2, dy1, dy2 = getx1x2y1y2(mainHeader['DATASEC'])
    #endregion: setups

    #!! main cycle
    for index in range(N):
        image = hduList[index].data
        header = hduList[index].header
        macro.cleanComments(header)
        #<== fit overscan and do trim
        if image is not None: #!! we have not data in the frame (should be the 0 frame, maybe..)
            # ysize  xsize
            nrows, ncolumns = image.shape
            if subDirection=='row':
                overscan = image[dy1-1:dy2, bx1-1:bx2] # see the top comments about the index
                data     = image[dy1-1:dy2, dx1-1:dx2]
                overscanMedian = np.median(overscan, axis=1)
                if fit: #!! fit overscan region
                    xx = np.arange(overscanMedian.size)
                    leg = Legendre.fit(xx, overscanMedian, 2)
                    overscanFit = leg(xx)
                    overscan = np.tile(overscanFit, (data.shape[1],1)).T
                    header.add_history('Model overscan by median and fit by legendre')
                else:
                    overscan = np.tile(overscanMedian, (data.shape[1],1)).T
                    header.add_history('Model overscan by median')
                data = data - overscan
                header.add_history('Substracted overscan on the row (along x) direction')
                header.add_history('Trimed to data section')
            else: #!! not tested below...
                overscan = image[by1-1:by2, dx1-1:dx2]
                data     = image[dy1-1:dy2, dx1-1:dx2]
                overscanMedian = np.median(overscan, axis=0)
                if fit:
                    xx = np.arange(overscanMedian.size)
                    leg = Legendre.fit(xx, overscanMedian, 2)
                    overscanFit = leg(xx)
                    overscan = np.tile(overscanFit, (data.shape[0],1))
                    header.add_history('Model overscan by median and fit by legendre')
                else:
                    overscan = np.tile(overscanMedian, (data.shape[0],1))
                    header.add_history('Model overscan by median')
                data = data - overscan
                header.add_history('Substracted overscan on the column (along y) direction')
                header.add_history('Trimed to data section')
        else: #!! data is None, may be the first frame, only change header
            data = None
            if subDirection=='row':
                if fit: #!! fit overscan region
                    header.add_history('Model overscan by median and fit by legendre')
                else:
                    header.add_history('Model overscan by median')
                header.add_history('Substracted overscan on the row (along x) direction')
                header.add_history('Trimed to data section')
            else: #!! not tested below...
                if fit:
                    header.add_history('Model overscan by median and fit by legendre')
                else:
                    header.add_history('Model overscan by median')
                header.add_history('Substracted overscan on the column (along y) direction')
                header.add_history('Trimed to data section')
        #==> fit overscan and do trim
        hduList[index].data = data
        hduList[index].header = header
    return hduList

def adu2electro(imageHDUlist, configs={}):#<==
    '''sub overscan and do trim using header imformation
    Parameters
    -------------------
    image : `str` or `astropy.io.fits.hdu.hdulist.HDUList`
        input image
    configs : `dict`, optional
        fit : weather to fit the overscan region
        subDirection : which way to (fit) and sub the overscan region
    Returns
    ------------------
    type : `astropy.io.fits.hdu.hdulist.HDUList`

    '''#==>
    #!! format input image
    exec(messager.initLoggerMacro)
    hduList = imageHDUlist
    N = len(hduList)

    #region: setups
    errorType = configs.get("errorType", "error")
    if errorType not in ['error', 'variance', 'ivar', None]:
        raise logger.Exception('unknown error type {}'.format(errorType))
    gainKey = configs.get("gainKey", 'EGAIN')
    noiseKey = configs.get("noiseKey", "ENOISE")

    mainHeader = hduList[0].header
    egain = mainHeader[gainKey]
    ereadnoise = mainHeader[noiseKey]
    #endregion: setups

    aux4 = []
    #!! main cycle
    for index in range(N):
        image = hduList[index].data
        header = hduList[index].header
        if image is None:
            header.add_history("ADU to Electron, add Var extension for each image")
            continue
        eimage = image * egain
        eheader = pyfits.header.Header()
        if errorType=='error':
            error = (eimage*(eimage>0) + ereadnoise**2)**0.5 #!! do the math
            header.add_history('ADU to electron, calculate error')
            header["EXTNAME"] = "SCI"
            eheader['EXTNAME'] = 'STD'
            aux4.append((eimage, header, error, eheader))
        elif errorType=='variance':
            error = (eimage*(eimage>0) + ereadnoise**2) #!! do the math
            header.add_history('ADU to electron, calculate variance')
            header["EXTNAME"] = "SCI"
            eheader['EXTNAME'] = 'VAR'
            aux4.append((eimage, header, error, eheader))
        elif errorType=='ivar':
            error = 1/(eimage*(eimage>0) + ereadnoise**2) #!! do the math
            header.add_history('ADU to electron, calculate ivar')
            header["EXTNAME"] = "SCI"
            eheader['EXTNAME'] = 'IVAR'
            aux4.append((eimage, header, error, eheader))
        else:
            header.add_history('ADU to electron, not calculate any error')
            header["EXTNAME"] = "SCI"
    primaryHDU = hduList[0]
    hduListList = []
    for eachAux4 in aux4:
        hduListList.append(pyfits.ImageHDU(eachAux4[0], eachAux4[1]))
        if len(eachAux4)==4:
            hduListList.append(pyfits.ImageHDU(eachAux4[2], eachAux4[3]))
    return pyfits.HDUList([primaryHDU] + hduListList)

def LACosmic(imageFile, configs={}):
    """input file name to process"""
    exec(messager.initLoggerMacro)
    hduList = pyfits.open(imageFile)
    N = len(hduList)

    #region: setups
    Ngrow   = configs.get('Ngrow', 1)
    #detection limit for cosmic rays (sigma)
    sigclip = configs.get('sigclip', 30.0)
    maxiter = configs.get('maxiter', 4)
    imgain  = configs.get('gain', 1)
    rdnoise=hduList["ENOISE"].data

    badFluxValue= configs.get('badFluxValue', "nan")
    if badFluxValue == 'inf':
        badFluxValue = np.inf
    elif badFluxValue == '-inf':
        badFluxValue = -np.inf
    elif badFluxValue == 'nan':
        badFluxValue = -np.nan

    badErrorValue=configs.get('badErrorValue', 'inf')
    if badErrorValue == 'inf':
        badErrorValue = np.inf
    elif badErrorValue == '-inf':
        badErrorValue = -np.inf
    elif badErrorValue == 'nan':
        badErrorValue = -np.nan
    #endregion: setups

    image = hduList["sci"].data
    error = hduList["std"].data
    header = hduList[0].header
    sheader = hduList["sci"].header
    eheader = hduList["std"].header
    cObj = la.cosmicsimage(image, gain=imgain,
               readnoise=rdnoise, sigclip = sigclip, noOutput=True)
    cObj.run(maxiter = maxiter, verbose=False) # max number of iterations
    # bad mask
    mask = cObj.mask.astype(np.short)
    assert mask.shape==image.shape
    ny, nx = mask.shape
    indcrs=np.where(mask==1);
    ncr=len(indcrs[0])
    logger.p("{} cosmic detected, growing pixels by {}".format(ncr, Ngrow))

    if Ngrow>0:
        iy = indcrs[0].astype(np.int32)
        ix = indcrs[1].astype(np.int32)
        iy_c = np.ctypeslib.as_ctypes(iy)
        ix_c = np.ctypeslib.as_ctypes(ix)
        mask_c = np.ctypeslib.as_ctypes(mask)
        shortClass = mask_c.__class__
        #void grow(short mask[],
        #          int y, int x,
        #          int* yIndexs, int* xIndexs,
        #          int indexLen, int grow) {
        grow_c.argtypes = [ shortClass, c_int, c_int, c_int_p, c_int_p, c_int, c_int]
        grow_c(mask_c, ny, nx, iy_c, ix_c, ncr, Ngrow)

    #if Ngrow>0:
        #for i in range(ncr):
            #ix=indcrs[0][i]
            #iy=indcrs[1][i]
            #mask[max(0,ix-Ngrow):min(nx-1, ix+Ngrow)+1, max(0, iy-Ngrow):min(ny-1, iy+Ngrow)+1]=1
        ##print("grow ok")

    image[mask==1]=badFluxValue
    hduList["sci"].data = image
    header.add_history('LACosmic: Maked Cosmic Rays with LA Cosmic and set bad flux pixel to {} with sigclip:{} and maxiter:{}'.format(badFluxValue, sigclip, maxiter))
    sheader.add_history('LACosmic: Maked Cosmic Rays with LA Cosmic and set bad flux pixel to {} with sigclip:{} and maxiter:{}'.format(badFluxValue, sigclip, maxiter))
    error[mask==1]=badErrorValue
    hduList["std"].data = error
    eheader.add_history('LACosmic: Maked Cosmic Rays with LA Cosmic and set bad flux pixel to {} with sigclip:{} and maxiter:{}'.format(badErrorValue, sigclip, maxiter))
    hduList.pop()
    return hduList

def growByMask(imageFile, mask, configs={}):
    exec(messager.initLoggerMacro)
    hduList = pyfits.open(imageFile)
    N = len(hduList)

    #region: setups
    Ngrow   = configs.get('Ngrow', 1)
    badFluxValue= configs.get('badFluxValue', "nan")
    if badFluxValue == 'inf':
        badFluxValue = np.inf
    elif badFluxValue == '-inf':
        badFluxValue = -np.inf
    elif badFluxValue == 'nan':
        badFluxValue = -np.nan

    badErrorValue=configs.get('badErrorValue', 'inf')
    if badErrorValue == 'inf':
        badErrorValue = np.inf
    elif badErrorValue == '-inf':
        badErrorValue = -np.inf
    elif badErrorValue == 'nan':
        badErrorValue = -np.nan
    #endregion: setups

    image = hduList["sci"].data
    error = hduList["std"].data
    header = hduList[0].header
    sheader = hduList["sci"].header
    eheader = hduList["std"].header

    mask = mask.copy()
    assert mask.shape==image.shape
    ny, nx = mask.shape
    indcrs=np.where(mask>0);
    ncr=len(indcrs[0])

    if Ngrow>0:
        iy = indcrs[0].astype(np.int32)
        ix = indcrs[1].astype(np.int32)
        iy_c = np.ctypeslib.as_ctypes(iy)
        ix_c = np.ctypeslib.as_ctypes(ix)
        mask_c = np.ctypeslib.as_ctypes(mask)
        shortClass = mask_c.__class__
        #void grow(short mask[],
        #          int y, int x,
        #          int* yIndexs, int* xIndexs,
        #          int indexLen, int grow) {
        grow_c.argtypes = [shortClass, c_int, c_int, c_int_p, c_int_p, c_int, c_int]
        grow_c(mask_c, ny, nx, iy_c, ix_c, ncr, Ngrow)

    image[mask>0]=badFluxValue
    hduList["sci"].data = image
    error[mask>0]=badErrorValue
    hduList["std"].data = error
    header.add_history("mask by aveCosmic")
    sheader.add_history("mask by aveCosmic")
    eheader.add_history("mask by aveCosmic")
    if len(hduList)==4:
        hduList.pop()
    return hduList



def combine(fileList, configs={}):#<==
    '''combine files
    '''#==>
    #region: setups
    exec(messager.initLoggerMacro)

    # setup input badFluxValue and badErrorValue
    badFluxValue= configs.get('badFluxValue', 'nan')
    if badFluxValue == 'inf':
        badFluxValue = np.inf
    elif badFluxValue == '-inf':
        badFluxValue = -np.inf
    elif badFluxValue == 'nan':
        badFluxValue = np.nan

    badErrorValue=configs.get('badErrorValue', 'inf')
    if badErrorValue == 'inf':
        badErrorValue = np.inf
    elif badErrorValue == '-inf':
        badErrorValue = -np.inf
    elif badErrorValue == 'nan':
        badErrorValue = np.nan

    method     = configs.get('method', 'mean')
    scaleByTime= configs.get('scaleByTime', 1)
    timeHeader= configs.get('timeHeader', 'EXPTIME')
    maxiter = configs.get("maxiter", 5)
    sigrej  = configs.get("sigrej")
    hsigrej  = configs.get("hsigrej")
    lsigrej  = configs.get("lsigrej")
    maxDev  = configs.get("maxDev", None)
    maxRej  = configs.get("maxRej", 1)
    percentileNorm = configs.get("percentileNorm", None)


    if method not in ['median', 'mean']:
        raise logger.Exception('unknown combine method')
    #endregion: setups

    #!! read file
    N = len(fileList)
    hduListList = [pyfits.open(each) for each in fileList]
    if scaleByTime: #!! normal by time
        exptime = [each[0].header["exptime"] for each in hduListList]
        for each, eachtime in zip(hduListList, exptime):
            each["SCI"].data *= (scaleByTime/eachtime)
            each["STD"].data *= (scaleByTime/eachtime)
            each["SCI"].header.add_history("scale from {}s to {}s".format(eachtime, scaleByTime))
            each["STD"].header.add_history("scale from {}s to {}s".format(eachtime, scaleByTime))
    if percentileNorm is not None:
        for each, eachtime in zip(hduListList, exptime):
            normFactor = np.nanpercentile(each["SCI"].data, percentileNorm)
            each["SCI"].data /= normFactor
            each["STD"].data /= normFactor
            each["SCI"].header.add_history("norm by divide percentile(SCI, {})".format(percentileNorm))
            each["STD"].header.add_history("norm by divide percentile(SCI, {})".format(percentileNorm))

    #if N==1:#!! only change some header info, scale to 1s  and return
        #hduList = hduListList[0]
        #header = hduList[0].header
        #sheader = hduList["sci"].header
        #eheader = hduList["std"].header
        #header["object"] = "combine @ {}s".format(scaleByTime)
        #header["exptime"] = scaleByTime
        #sheader["object"] = "combine sci@ {}s".format(scaleByTime)
        #sheader["exptime"] = scaleByTime
        #eheader["object"] = "combine std@ {}s".format(scaleByTime)
        #eheader["exptime"] = scaleByTime
        #header.add_history("combine using single file {}".\
                            #format(os.path.basename(fileList[0])))
        #sheader.add_history("combine using single file {}".\
                            #format(os.path.basename(fileList[0])))
        #eheader.add_history("combine using single file {}".\
                            #format(os.path.basename(fileList[0])))
        #return hduList

    #!! below are for N>1
    fileNameList = [os.path.basename(each) for each in fileList]
    objectList = [each[0].header["object"] for each in hduListList]

    header = pyfits.Header()
    header["OBJECT"] = "combine of {}".format(",".join(fileNameList))
    header[timeHeader] = scaleByTime

    sheader = pyfits.Header()
    sheader[timeHeader] = scaleByTime
    sheader["extname"] = "SCI"
    sheader["OBJECT"] = "combine of {}".format(fileNameList)

    eheader = pyfits.Header()
    eheader[timeHeader] = scaleByTime
    eheader["extname"] = "STD"
    eheader["OBJECT"] = "combine of {}".format(fileNameList)

    gheader = pyfits.Header()
    gheader[timeHeader] = scaleByTime
    gheader["extname"] = "GOODNUM"
    gheader["OBJECT"] = "combine of {}".format(fileNameList)

    header.add_history("combine files")
    for eachName,eachObject in zip(fileNameList, objectList):
        header.add_history("{} {}".format(eachName, eachObject))

    primaryHDU = pyfits.PrimaryHDU(None, header=header)
    a,(y,x) = N, hduListList[0]["SCI"].shape
    data = np.empty((a, y, x), dtype=np.float64)
    ivar = np.empty((a, y, x), dtype=np.float64)
    for i, eachObj in enumerate(hduListList):
        data[i,:,:] = eachObj["sci"].data[:,:]
        ivar[i,:,:] = 1/(eachObj["STD"].data[:,:])**2

    if method=='mean':
        if N>1:
            result = astroMath.iterstat3dC(data, invvar = ivar,
                                           maxiter=maxiter, sigrej=sigrej, hsigrej=hsigrej, lsigrej=lsigrej,
                                           maxDev=maxDev, maxRej=maxRej)
            mean = result["mean"]
            badmask = result["badmask"] #!! 3D
            goodmask = np.logical_not(badmask) #!! 3D
            newBadMask = result["newBadMask"] #!! 3D
            ivar[badmask] = 0 #!! 3D
            newstd = 1/np.sqrt(ivar.sum(axis=0)) #!! 2D
            goodNumber = goodmask.sum(axis=0) #!! 2D
        else:
            mean = data[0]
            badmask = ivar==0
            goodmask = np.logical_not(badmask)
            newBadMask = badmask
            ivar[badmask] = 0
            newstd = 1/np.sqrt(ivar.sum(axis=0))
            goodNumber = goodmask.sum(axis=0)

        dataHDU = pyfits.ImageHDU(mean, header = sheader)
        errorHDU = pyfits.ImageHDU(newstd, header = eheader)
        goodNumberHDU = pyfits.ImageHDU(goodNumber, header=gheader)
        workList = [primaryHDU, dataHDU, errorHDU, goodNumberHDU]
        debugList = []
        for i, (eachObj,eachFile) in enumerate(zip(hduListList, fileNameList)):
            aux1 = eachObj["SCI"]
            aux1.header["extname"] = eachFile+"-debug"
            aux2 = eachObj["STD"]
            aux2.header["extname"] = eachFile+"-STD-debug"
            nHeader = pyfits.Header()
            nHeader["extname"] = eachFile+"-NewBadPixel-debug"
            nHeader["object"] = aux1.header["object"]
            aux3 = pyfits.ImageHDU(newBadMask[i].astype(np.short), nHeader)
            debugList.extend([aux1, aux2, aux3])

        hduList = pyfits.HDUList(workList+debugList)
        return hduList

def scaleCombine(fileList, outfile='', configs={}):
    """scale it by flux sum and combine, only use for twilight"""
   # setups
    # setup input badFluxValue and badErrorValue
    badFluxValue= configs.get('badFluxValue', -65535)
    if badFluxValue == 'inf':
        badFluxValue = np.inf
    elif badFluxValue == '-inf':
        badFluxValue = -np.inf
    elif badFluxValue == 'nan':
        badFluxValue = -np.nan

    badErrorValue=configs.get('badErrorValue', 'inf')
    if badErrorValue == 'inf':
        badErrorValue = np.inf
    elif badErrorValue == '-inf':
        badErrorValue = -np.inf
    elif badErrorValue == 'nan':
        badErrorValue = -np.nan

    errorIndex = configs.get('errorIndex', 1)
    # the python code to getMask
    getMask    = configs.get('getMask', '')
    method     = configs.get('method', 'median')
    timeHeader= configs.get('timeHeader', 'EXPTIME')
    # read file
    N = len(fileList)
    filenames = configs.get("filenames", ["?"] * N)
    headerList = [pyfits.Header() for i in range(N)]
    first = fileList.pop()
    thisObj = pyfits.open(first)
    thisData  = thisObj[0].data
    thisHeader = thisObj[0].header
    headerList[0]["COMMENT"] = filenames[0]
    headerList[0]["EXPTIME"] = thisHeader["EXPTIME"]
    thisMed = np.nanpercentile(thisData, 80)
    ny, nx = thisData.shape
    thisDataH = thisObj[0].header
    dataHeader = thisObj[0].header
    if N==1:
        raise Exception('only one image, no need to combine')
    dataStackArray = np.zeros((N, ny, nx))
    thisTime = thisDataH[timeHeader]
    firstData = thisData/thisTime/thisMed
    del dataHeader[timeHeader]
    dataHeader[timeHeader] = 1
    thisError = thisObj[errorIndex].data
    errorHeader = thisObj[errorIndex].header
    errorStackArray= np.zeros((N, ny, nx))
    # mask is bad mask
    thisMask=~np.isfinite(thisError)
    maskStackArray = np.zeros((N, ny, nx))>0
    maskStackArray[0,:,:] = thisMask
    thisData[thisMask]=np.nan
    dataStackArray[0,:,:] = thisData/thisTime/thisMed
    errorStackArray[0,:,:] = (thisError/thisTime/thisMed)**2
    # row column
    for i,each in enumerate(fileList):
        thisObj = pyfits.open(each)
        thisData  = thisObj[0].data
        thisHeader = thisObj[0].header
        headerList[i+1]["COMMENT"] = filenames[i+1]
        headerList[i+1]["EXPTIME"] = thisHeader["EXPTIME"]
        thisError = thisObj[errorIndex].data
        thisMask=~np.isfinite(thisError)
        maskStackArray[i+1,:,:] = thisMask
        thisData[thisMask]=np.nan
        # make scale
        thisFactor = np.nansum(firstData*thisData/thisError**2)/np.nansum(thisData**2/thisError**2)
        dataStackArray[i+1,:,:] = thisData*thisFactor
        errorStackArray[i+1,:,:] = (thisError*thisFactor)**2
   # combine all good pixel using reject mean
   # first cal
    # still bad mask after the combine
    goodPixCount = N-np.nansum(maskStackArray, axis=0)
    #dataStackArray[0][allBadMask] = badFluxValue
    dataCombined = np.nanmean(dataStackArray, axis=0)
    stdCombined  = np.nanstd(dataStackArray, axis=0)
    rejMask = np.abs(dataStackArray - dataCombined)>(2.5*stdCombined)
    maskStackArray[rejMask]=True
   # second cal
    dataStackArray[maskStackArray] = np.nan
    # still bad mask after the combine
    goodPixCount = N-np.nansum(maskStackArray, axis=0)
    dataCombined = np.nanmean(dataStackArray, axis=0)
    allBadMask = goodPixCount == 0

    dataCombined[allBadMask] = badFluxValue
    dataHeader['HISTORY']= 'combine using mean, bad pixel set to {}'.format(badFluxValue)
    errorHeader['HISTORY']= \
        'combine using mean, bad error set to {}'.format(badErrorValue)
    errorStackArray[maskStackArray] = np.nan
    errorCombined = np.sqrt(np.nansum(errorStackArray, axis=0))/goodPixCount
    errorCombined[allBadMask] = badErrorValue
    if outfile:
        outfile = outfile.split(".fits")[0]+"CombineDebug.fits"
        dataList = []
        dataList.append(pyfits.PrimaryHDU(dataCombined, header=dataHeader))
        #if withError:
        #    dataList.append(pyfits.ImageHDU(errorCombined, header=errorHeader))
        #dataList.append(pyfits.ImageHDU(allBadMask.astype(np.int32)))
        for i in range(dataStackArray.shape[0]):
            dataList.append(pyfits.ImageHDU(dataStackArray[i], header=headerList[i]))
            #if withError:
            #    dataList.append(pyfits.ImageHDU(errorStackArray[i], header=errorHeader))
            #dataList.append(pyfits.ImageHDU(maskStackArray[i].astype(np.int32)))
        fits = pyfits.HDUList(dataList)
        fits.writeto(outfile, clobber='True')
    return dataCombined, dataHeader, errorCombined, errorHeader, allBadMask

def interpImage(image, badMask, configs={}):
    exec(messager.initLoggerMacro)

    assert image.shape==badMask.shape
    kind = configs.get('kind','linear')
    axis = configs.get('axis', "column")
    ignoreWarnings = configs.get("ignoreWarnings", [])
    if axis not in ["column", "row"]: raise logger.Exception("unknown Parameters axis:{}".format(axis))
    if axis=="column":
        image = image.T
        badMask = badMask.T

    ny, nx = image.shape
    yy, xx = np.where(badMask)
    yy = np.unique(yy)
    xCoors = np.arange(nx,dtype=np.int)
    for row in yy:
        good = np.logical_not(badMask[row])
        thisX = xCoors[good]
        thisY = image[row][good]
        #fit = interp.UnivariateSpline(thisX, thisY)
        if good.sum()<2:
            if row not in ignoreWarnings:
                logger.warn("fiber {} all bad".format(row+1))
        else:
            fit = interp.interp1d(thisX, thisY, kind=kind, fill_value="extrapolate")
            image[xind, :] = fit(xCoors)

    if axis=="column":
        image = image.T
        badMask = badMask.T

    return image

def subtract(imageA, errorA, imageB, errorB, configs={}):
    assert imageA.shape==imageB.shape
    assert errorA.shape==imageA.shape
    assert errorB.shape==imageB.shape
    maskA = ~np.isfinite(errorA)
    maskB = ~np.isfinite(errorB)
    mask = np.logical_and(maskA, maskB)
    image = imageA - imageB
    image[mask] = np.nan
    error = np.sqrt(errorA**2 + errorB**2)
    return image, error

def fixpixel(imageFile, configs={}):
    logger = configs["logger"]
    direction = configs.get("direction")
    if direction is None:
        direction = "row"
    elif direction not in ["row", "column"]:
        raise logger.Exception("interpolate direction should be either row or column")
    fits = pyfits.open(imageFile)[:3]
    fits.extend(pyfits.open(imageFile)[1:3])
    fits[3].header["object"] = "origin sci"
    fits[3].header["EXTNAME"] = "OSCI"
    fits[4].header["object"] = "origin std"
    fits[4].header["EXTNAME"] = "OSTD"
    data = fits[1].data
    std = fits[2].data
    if direction == "column":
        data = data.T
        std = std.T

    linesTodo = np.unique(np.where(np.logical_not(np.isfinite(data)))[0])
    xx = np.arange(data.shape[1])
    from .. import de
    for each in linesTodo:
        result = astroMath.linearInterpWithError(xx, data[each], std[each], xx)
        data[each, :] = result["y"][:]
        std[each, :] = result["sigma"][:]

    if direction == "column":
        data = data.T
        std = std.T

    return fits

def toFits(filename, dataHeaderPair):
    first = dataHeaderPair.pop(0)
    assert len(first)==2, "the second input should have format (data, header)"
    data0 = pyfits.PrimaryHDU(first[0], header=first[1])
    allData = [data0]
    for each in dataHeaderPair:
        assert len(each)==2, "the second input should have format (data, header)"
        allData.append(pyfits.ImageHDU(each[0], header=each[1]))
    fits  = pyfits.HDUList(allData)
    fits.writeto(filename, clobber="True")

def addToFits(filename, dataHeaderPair):
    obj = pyfits.open(filename)
    for each in dataHeaderPair:
        obj.append(pyfits.ImageHDU(each[0], header=each[1]))
    obj.writeto(filename, clobber="True")




#===============================================================
# backup
#def LACosmic(imageFile, configs={}):
    #"""input file name to process"""
    #exec(messager.initLoggerMacro)
    #hduList = pyfits.open(imageFile)
    #N = len(hduList)

    ##region: setups
    #Ngrow   = configs.get('Ngrow', 1.0)
    ##detection limit for cosmic rays (sigma)
    #sigclip = configs.get('sigclip', 30.0)
    #maxiter = configs.get('maxiter', 4)
    #imgain  = configs.get('gain', 1)
    #rdnoise=hduList[0].header['ENOISE']

    #badFluxValue= configs.get('badFluxValue', "nan")
    #if badFluxValue == 'inf':
        #badFluxValue = np.inf
    #elif badFluxValue == '-inf':
        #badFluxValue = -np.inf
    #elif badFluxValue == 'nan':
        #badFluxValue = -np.nan

    #badErrorValue=configs.get('badErrorValue', 'inf')
    #if badErrorValue == 'inf':
        #badErrorValue = np.inf
    #elif badErrorValue == '-inf':
        #badErrorValue = -np.inf
    #elif badErrorValue == 'nan':
        #badErrorValue = -np.nan
    ##endregion: setups

    #image = hduList["sci"].data
    #error = hduList["std"].data
    #header = hduList[0].header
    #sheader = hduList["sci"].header
    #eheader = hduList["std"].header
    #cObj = la.cosmicsimage(image, gain=imgain,
               #readnoise=rdnoise, sigclip = sigclip, noOutput=True)
    #cObj.run(maxiter = maxiter, verbose=False) # max number of iterations
    ## bad mask
    #mask = cObj.mask.astype(np.short)
    #assert mask.shape==image.shape
    #nx, ny = mask.shape
    #indcrs=np.where(mask==1);
    #ncr=len(indcrs[0])
    #logger.p("{} cosmic detected, growing pixels by {}".format(ncr, Ngrow))
    #if Ngrow>0:
        #for i in range(ncr):
            #ix=indcrs[0][i]
            #iy=indcrs[1][i]
            #mask[max(0,ix-Ngrow):min(nx-1, ix+Ngrow)+1, max(0, iy-Ngrow):min(ny-1, iy+Ngrow)+1]=1
        ##print("grow ok")
    #image[mask==1]=badFluxValue
    #hduList["sci"].data = image
    #header.add_history('LACosmic: Maked Cosmic Rays with LA Cosmic and set bad flux pixel to {}'.format(badFluxValue))
    #sheader.add_history('LACosmic: Maked Cosmic Rays with LA Cosmic and set bad flux pixel to {}'.format(badFluxValue))
    #error[mask==1]=badErrorValue
    #hduList["std"].data = error
    #eheader.add_history('LACosmic: Maked Cosmic Rays with LA Cosmic and set bad flux pixel to {}'.format(badErrorValue))
    #return hduList
#def multiFileCosmic(imageFileList, configs={}):
    ##!! format input image
    #exec(messager.initLoggerMacro)
    #hduListList = [pyfits.open(eachFile) for eachFile in imageFileList]

    ##region: setups
    #sigclip = configs.get('sigclip', 30.0)
    #maxiter = configs.get('maxiter', 4)

    #badFluxValue= configs.get('badFluxValue', "nan")
    #if badFluxValue == 'inf':
        #badFluxValue = np.inf
    #elif badFluxValue == '-inf':
        #badFluxValue = -np.inf
    #elif badFluxValue == 'nan':
        #badFluxValue = -np.nan

    #badErrorValue=configs.get('badErrorValue', 'inf')
    #if badErrorValue == 'inf':
        #badErrorValue = np.inf
    #elif badErrorValue == '-inf':
        #badErrorValue = -np.inf
    #elif badErrorValue == 'nan':
        #badErrorValue = -np.nan
    ##endregion: setups

    #image = hduList["sci"].data
    #error = hduList["std"].data
    #header = hduList[0].header
    #sheader = hduList["sci"].header
    #eheader = hduList["var"].header
    #cObj = LACosmic.cosmicsimage(image, gain=imgain,
                             #readnoise=rdnoise, sigclip = sigclip, noOutput=True)
    #cObj.run(maxiter = maxiter, verbose=False) # max number of iterations
    ## bad mask
    #mask = cObj.mask.astype(np.short)
    #assert mask.shape==image.shape
    #ny, nx = mask.shape
    #indcrs=np.where(mask==1)
    #ncr=len(indcrs[0])
    #logger.p("{} cosmic detected, growing pixels by {}".format(ncr, Ngrow))

    #if Ngrow>0:
        #iy = indcrs[0].astype(np.int)
        #ix = indcrs[1].astype(np.int)
        #iy_c = np.ctypeslib.as_ctypes(iy)
        #ix_c = np.ctypeslib.as_ctypes(ix)
        #mask_c = np.ctypeslib.as_ctypes(mask)
        ##void grow(short mask[],
        ##          int y, int x,
        ##          int* yIndexs, int* xIndexs,
        ##          int indexLen, int grow) {
        #grow_c.argtypes = [ c_short_p, c_int, c_int, c_int_p, c_int_p, c_int, c_int]
        #grow_c(mask_c, ny, nx, iy_c, ix_c, ncr, Ngrow)

    #image[mask==1]=badFluxValue
    #hduList["sci"].data = image
    #header.add_history('multiFileCosmic: Maked Cosmic Rays with MultiFileCosmic and set bad flux pixel to {}'.format(badFluxValue))
    #sheader.add_history('multiFileCosmic: Maked Cosmic Rays with MultiFileCosmic and set bad flux pixel to {}'.format(badFluxValue))
    #error[mask==1]=badErrorValue
    #hduList["std"].data = error
    #eheader.add_history('multiFileCosmic: Maked Cosmic Rays with MultiFileCosmic and set bad flux pixel to {}'.format(badErrorValue))
    #return hduList
#def combine(fileList, configs={}):
    ##region: setups
    #exec(messager.initLoggerMacro)

    ## setup input badFluxValue and badErrorValue
    #badFluxValue= configs.get('badFluxValue', 'nan')
    #if badFluxValue == 'inf':
        #badFluxValue = np.inf
    #elif badFluxValue == '-inf':
        #badFluxValue = -np.inf
    #elif badFluxValue == 'nan':
        #badFluxValue = np.nan

    #badErrorValue=configs.get('badErrorValue', 'inf')
    #if badErrorValue == 'inf':
        #badErrorValue = np.inf
    #elif badErrorValue == '-inf':
        #badErrorValue = -np.inf
    #elif badErrorValue == 'nan':
        #badErrorValue = np.nan

    #withError  = configs.get('withError',  1)
    ## the error image index in the input
    #errorIndex = configs.get('errorIndex', 1)
    ## whether the input have a mask image
    #withMask   = configs.get('withMask',   1)
    ## the python code to getMask
    #getMask    = configs.get('getMask',
            #'thisMask=~np.isfinite(thisError)')
    #method     = configs.get('method', 'median')
    #scaleByTime= configs.get('scaleByTime', 1)
    #timeHeader= configs.get('timeHeader', 'EXPTIME')
    #if method not in ['median', 'mean']:
        #raise Exception('unknown combine method')
    ##endregion: setups
    ##!! read file
    #N = len(fileList)
    #filenames = configs.get("filenames", ["?"] * N)
    #headerList = [pyfits.Header() for i in range(N)]
    #first = fileList.pop()
    #thisObj = pyfits.open(first)
    #thisData  = thisObj[0].data
    #headerList[0]["COMMENT"] = filenames[0]
    #ny, nx = thisData.shape
    #thisDataH = thisObj[0].header
    #dataHeader = thisObj[0].header
    #if N==1:
        #raise Exception('only one image, no need to combine')
    #if scaleByTime:
        #dataStackArray = np.zeros((N, ny, nx))
        #thisTime = thisDataH[timeHeader]
        #headerList[0]["EXPTIME"] = thisTime
        #dataStackArray[0,:,:] = thisData/thisTime
        #del dataHeader[timeHeader]
        #dataHeader[timeHeader] = 1
        #if withError:
            #thisError = thisObj[errorIndex].data
            #errorHeader = thisObj[errorIndex].header
            #errorStackArray= np.zeros((N, ny, nx))
            #errorStackArray[0,:,:] = (thisError/thisTime)**2
        ## mask is bad mask
        #if withMask:
            #exec(getMask)
            #maskStackArray = np.zeros((N, ny, nx))>0
            #maskStackArray[0,:,:] = thisMask
        ## row column
        #for i,each in enumerate(fileList):
            #thisObj = pyfits.open(each)
            #thisData  = thisObj[0].data
            #thisDataH = thisObj[0].header
            #headerList[i+1]["COMMENT"] = filenames[i+1]
            #thisTime = thisDataH[timeHeader]
            #headerList[i+1]["EXPTIME"] = thisTime
            #dataStackArray[i+1,:,:] = thisData/thisTime
            #if withError:
                #thisError = thisObj[errorIndex].data
                #thisErrorH= thisObj[errorIndex].header
                #errorStackArray[i+1,:,:] = (thisError/thisTime)**2
            #if withMask:
                #exec(getMask)
                #maskStackArray[i+1,:,:] = thisMask
    #else:
        #dataStackArray = np.zeros((N, ny, nx))
        #dataStackArray[0,:,:] = thisData
        #if withError:
            #thisError = thisObj[errorIndex].data
            #errorStackArray= np.zeros((N, ny, nx))
            #errorStackArray[0,:,:] = (thisError)**2
        #if withMask:
            #exec(getMask)
            #maskStackArray = np.zeros((N, ny, nx),dtype=np.int32)>0
            #maskStackArray[0,:,:] = thisMask
        ## row column
        #for i,each in enumerate(fileList):
            #thisObj = pyfits.open(each)
            #thisData  = thisObj[0].data
            #headerList[i+1]["COMMENT"] = filenames[i+1]
            #dataStackArray[i+1,:,:] = thisData
            #if withError:
                #thisError = thisObj[errorIndex].data
                #errorStackArray[i+1,:,:] = (thisError)**2
            #if withMask:
                ## get mask use "getMask"
                #exec(getMask)
                #maskStackArray[i+1,:,:] = thisMask
    ## combine all good pixel using different method
    #if not withMask:
        #maskStackArray= np.zeros((N, ny, nx))>0
    #dataStackArray[maskStackArray] = np.nan
    ## still bad mask after the combine
    #goodPixCount = N-np.nansum(maskStackArray, axis=0)
    #allBadMask = goodPixCount == 0
    #xx, yy = np.meshgrid(range(nx), range(ny))
    ##dataStackArray[0][allBadMask] = badFluxValue
    ##pdb.set_trace()
    #if method=='mean':
        #dataCombined = np.nanmean(dataStackArray, axis=0)
        #dataCombined[allBadMask] = badFluxValue
        #dataHeader['HISTORY']= 'combine using mean, bad pixel set to {}'.format(badFluxValue)
        #if withError:
            #errorHeader['HISTORY']=\
                    #'combine using mean, bad error set to {}'.format(badErrorValue)
            #errorStackArray[maskStackArray] = np.nan
            #errorCombined = np.sqrt(np.nansum(errorStackArray, axis=0))/goodPixCount
            #errorCombined[allBadMask] = badErrorValue
        #if outfile:
            #outfile = outfile.split(".fits")[0]+"CombineDebug.fits"
            #dataList = []
            #dataList.append(pyfits.PrimaryHDU(dataCombined, header=dataHeader))
            ##if withError:
            ##    dataList.append(pyfits.ImageHDU(errorCombined, header=errorHeader))
            ##dataList.append(pyfits.ImageHDU(allBadMask.astype(np.int32)))
            #for i in range(dataStackArray.shape[0]):
                #dataList.append(pyfits.ImageHDU(dataStackArray[i], header=headerList[i]))
                ##if withError:
                ##    dataList.append(pyfits.ImageHDU(errorStackArray[i], header=errorHeader))
                ##dataList.append(pyfits.ImageHDU(maskStackArray[i].astype(np.int32)))
            #fits = pyfits.HDUList(dataList)
            #fits.writeto(outfile, clobber='True')
        #if withError:
            #return dataCombined, dataHeader, errorCombined, errorHeader, allBadMask
        #else:
            #return dataCombined, dataHeader,  allBadMask
    #else: # median
        #tempMask = np.array([allBadMask]*N)
        #dataStackArray[tempMask] = badFluxValue
        #dataCombined = np.nanmedian(dataStackArray, axis=0)
        #diff = np.abs(dataCombined - dataStackArray)
        #indexs = np.nanargmin(diff, axis=0)
        #dataCombined[allBadMask] = badFluxValue
        #dataHeader['HISTORY']= 'combine using median, bad pixel set to {}'.format(badFluxValue)
        #if withError:
            #errorHeader['HISTORY']=\
                    #'combine using median, bad error set to {}'.format(badErrorValue)
            #errorStackArray[maskStackArray] = np.nan
            #errorCombined = errorStackArray[indexs, yy, xx]
            #errorCombined[allBadMask] = badErrorValue
        #if outfile:
            #outfile = outfile.split(".fits")[0]+"CombineDebug.fits"
            #dataList = []
            #dataList.append(pyfits.PrimaryHDU(dataCombined, header=dataHeader))
            ##if withError:
            ##    dataList.append(pyfits.ImageHDU(errorCombined, header=errorHeader))
            ##dataList.append(pyfits.ImageHDU(allBadMask.astype(np.int32)))
            #for i in range(dataStackArray.shape[0]):
                #dataList.append(pyfits.ImageHDU(dataStackArray[i], header=headerList[i]))
                ##if withError:
                ##    dataList.append(pyfits.ImageHDU(errorStackArray[i], header=errorHeader))
                ##dataList.append(pyfits.ImageHDU(maskStackArray[i].astype(np.int32)))
            #fits = pyfits.HDUList(dataList)
            #fits.writeto(outfile, clobber='True')
        #if withError:
            #return dataCombined, dataHeader, errorCombined, errorHeader, allBadMask
        #else:
            #return dataCombined, dataHeader,  allBadMask
